#
# AUTHOR:	Danny Wynne
#		 	wynne100@gmail.com
#			www.dannywynne.com/blog
#
"""
Quickly open files with a Mac OS-like spotlight tool.	 

DESCRIPTION: 
	Overlays a text entry field in the UI that allows you to quickly open files in your project and 
custom directories. The tool indexes all files of a specified extension and autocompletes as you type.
It will autocomplete to the items you open most first. It also automatically organizes any versioned files
so that the latest version is on top.
Tested in maya 2011 and 2012 on windows

USAGE:
	place the script in your maya scripts directory. Run the python code:

import maya_spotlight as sp
sp.main()

press ctrl+p to pop up the tool and begin typing. 
you can add additional directories below in the settings section
"""
from PyQt4 import QtGui, QtCore
import sip
import maya.OpenMayaUI as mui
import maya.cmds as cmds
import maya.mel as mel
import sqlite3 as sql
import re, os, random, functools
import cPickle as pickle

#SETTINGS
#all the directories that the tool will search
directories = [
	]
#dont search folders with this names
exclude_folders = [
	"images",
	"renderData",
	"particles",
	]
#index only these types of files:
extensions = [
	".ma",
	".mb",
]
#your maya user apps directory:
maya_app_dir = os.getenv("MAYA_APP_DIR")
#END SETTINGS

trailing_numbers_re = re.compile('[0-9]+$')

def maya_window():
	ptr = mui.MQtUtil.mainWindow()
	return sip.wrapinstance(long(ptr), QtCore.QObject)

class MayaSpotlight(QtGui.QSplashScreen):
	"""Qt UI overlay with a autocomplete text entry and list of search results."""	
	def __init__(self, parent = None):
		super(QtGui.QSplashScreen, self).__init__(parent)
		self.setGeometry(100,100, 800,200)
		self.setWindowOpacity(.7)

		self.layout = QtGui.QVBoxLayout()
		self.setLayout(self.layout)
		self.le = CustomLineEdit(self);self.layout.addWidget(self.le)
		self.le.focusOut.connect(self.hide)
		self.le.setStyleSheet("background:opaque;")
		self.lw = QtGui.QListWidget(self);self.layout.addWidget(self.lw)
		self.lw.setFocusPolicy(0)
		self.lw.setStyleSheet("background:opaque;")

		self.le.autoComplete.connect(self.auto_complete)
		self.le.up.connect(functools.partial(self.list_nav, -1))
		self.le.down.connect(functools.partial(self.list_nav, 1))
		self.le.cancel.connect(self.cancel)
		self.le.returnPressed.connect(self.return_pressed)

		self.lw.itemSelectionChanged.connect(self.item_selection_changed)

		#private members
		directories.append(os.path.normpath(cmds.workspace(q=True, rd=True)))
		self.dir_cache = cache_dir_tree(directories, extensions, exclude_folders)
		self.current_text = ""

	def cancel(self):
		self.le.clearFocus()

	def return_pressed(self):
		selected_items = self.lw.selectedItems()
		if len(selected_items) == 0: return
		cache = selected_items[0].cache
		filename_id = os.path.join(cache[3], cache[4])
		self.le.clearFocus()
		record_hit(filename_id)
		print os.path.join(cache[3], cache[2])
		cmds.file(os.path.join(cache[3], cache[2]), open=True, f=1)

	def item_selection_changed(self):
		selected_items = self.lw.selectedItems()
		if len(selected_items) == 0: return
		cache = selected_items[0].cache
		self.set_auto_complete_text(self.current_text, cache[2])

	def list_nav(self, direction):
		next_row = (self.lw.currentRow()+direction)%self.lw.count()
		self.lw.setCurrentRow(next_row)

	def auto_complete(self):
		self.lw.clear()
		text = str(self.le.text())
		self.current_text = text
		if len(text) < 1: 
			return
		dir_cache_search = []
		for r in self.dir_cache:
			if r[2].find(text)==0:
				dir_cache_search.append(r)
		if dir_cache_search:
			self.set_auto_complete_text(text, dir_cache_search[0][2])
			for c in dir_cache_search[:10]:
				full_path = os.path.join(c[3], c[2])
				item = QtGui.QListWidgetItem(full_path)
				item.cache = c
				self.lw.addItem(item)
			self.lw.setCurrentRow(0)
			
	def set_auto_complete_text(self, original_text, auto_text):
		auto_text = auto_text.partition(original_text)[2]
		self.le.setText(original_text+auto_text)
		self.le.setSelection(len(original_text+auto_text), -len(auto_text))

class CustomLineEdit(QtGui.QLineEdit):
	"""Line edit with additional signals"""
	focusOut = QtCore.pyqtSignal()
	autoComplete = QtCore.pyqtSignal()
	up = QtCore.pyqtSignal()
	down = QtCore.pyqtSignal()
	cancel = QtCore.pyqtSignal()
	def focusOutEvent(self, event):
		self.focusOut.emit()
	def keyPressEvent(self, event):
		if event.key() == QtCore.Qt.Key_Backspace:
			if self.hasSelectedText():
				self.backspace()
			self.backspace()
			self.autoComplete.emit()
		elif event.key() == QtCore.Qt.Key_Up:
			self.up.emit()
		elif event.key() == QtCore.Qt.Key_Escape:
			self.cancel.emit()
		elif event.key() == QtCore.Qt.Key_Down:
			self.down.emit()
		else:
			super(CustomLineEdit, self).keyPressEvent(event)
			self.autoComplete.emit()

def cache_dir_tree(directories, include_ext, exclude_folders = None):
	"""walks directories and caches the filenames in a list

	The cached filenames are each lists of relevant information about the file.
	The cache entry includes the file version (determined by trailing numbers), how
	many times the file has been accessed by the spotlight tool, the filename and the
	directory.
	Args:
		directores(list) - list of directores to walk and index/cache
		include_ext(list) - list of extensions to index. [".ma", ".mb"]
		exclude_folders(list) - list of folders to exclude (ie no need to search images
			for .ma files) 
	Returns:
		cached_dir(list) - a list of lists. [[hits(int), version(int), filename(str), 
			dirpath(str), versionless_name(str)],]
			hits - how often the file has been accessed. versionless_name - name of the file with 
			version number striped out, this is used to uniquely identify groups of versioned files
			so that accessing one adds a hit for all.
	"""

	cached_dir = []
	searched_dirs = set()
	def _find(x_str, y_list):
		x_str_len = len(x_str)
		for y in y_list: 
			if x_str.find(y)+len(y) == x_str_len: return True
	for dir in directories:
		for dirpath, dirnames, filenames in os.walk(dir):
			if dirpath in searched_dirs: continue
			searched_dirs.add(dirpath)
			if exclude_folders:
				dirnames[:] = [d for d in dirnames if not d in exclude_folders]
			for filename in filenames:
				if _find(filename, include_ext):
					version, versionless_name = version_from_filename(filename)
					hits = get_hits(os.path.join(dirpath, versionless_name))
					row = [hits, version, filename, dirpath, versionless_name]
					cached_dir.append(row)
	cached_dir.sort(key=lambda x: x[1])
	cached_dir.reverse()
	cached_dir.sort(key=lambda x: os.path.join(x[3], x[4]))
	cached_dir.reverse()
	cached_dir.sort(key=lambda x: x[0])
	cached_dir.reverse()
	return cached_dir

def rreplace(s, old, new, count=1):
	return (s[::-1].replace(old[::-1], new[::-1], count))[::-1]
	
def version_from_filename(filename):
	"""gets the version number from a file name
	
	Args:
		filename(str)
	
	Returns:
	    version(int)
	    versionless_name(str) - name of the file with version number replaced with "?"
	"""
	_rp = filename.rpartition(".")
	basename, ext = _rp[0], "."+_rp[2]
	if not basename:
		basename, ext = filename, ""
	version_str = trailing_numbers_re.findall(basename)
	if version_str:
		version, versionless_name = int(version_str[-1]), rreplace(basename, version_str[-1], "?", 1)+ext
	else:
		version, versionless_name = 0, basename
	return version, versionless_name

def record_hit(filename_id):
	hit_records = unpickle_data("mayaspotlight.data")
	if not hit_records:
		hit_records = {}
	if filename_id in hit_records:
		hit_records[filename_id] += 1
	else:
		hit_records[filename_id] = 1
	pickle_data(hit_records, "mayaspotlight.data")

def get_hits(filename_id):
	hit_records = unpickle_data("mayaspotlight.data")
	if not hit_records: return 0
	if filename_id in hit_records:
		return hit_records[filename_id]
	else:
		return 0

def pickle_data(data, name, path = maya_app_dir):
	if not os.path.exists(path): 
		os.makedirs(path)
	file_object = open(os.path.join(path, name), "w")
	pickle.dump(data, file_object)
	file_object.close()
	
def unpickle_data(name, path = maya_app_dir):
	filepath = os.path.join(path, name)
	if os.path.isfile(filepath):
		file_object = open(filepath, "r")
		pickle_data = pickle.load(file_object)
		file_object.close()
		return pickle_data
	return None

class MayaSpotlightShortcuts(QtCore.QObject):
	"""event filter to be installed on maya main window. Shortcut is ctrl+p"""
	start = QtCore.pyqtSignal()
	def __init__(self):
		super(MayaSpotlightShortcuts, self).__init__()
	def eventFilter(self, obj, event):
		if event.type() == QtCore.QEvent.KeyPress:
			if event.key() == 80 and event.modifiers() == QtCore.Qt.ControlModifier:
				self.start.emit()
				return True
		return False

def install_shortcut():		
	global spotlight_shortcuts
	qt_maya_window = maya_window()
	spotlight_shortcuts = MayaSpotlightShortcuts()
	try:
		qt_maya_window.removeEventFilter(spotlight_shortcuts)
	except:
		pass
	qt_maya_window.installEventFilter(spotlight_shortcuts)
	return spotlight_shortcuts

def do_spotlight():
	"""instance MayaSpotlight, fit it nicely to the maya main window, and set focus"""
	global spotlight
	try:
		spotlight.hide()
	except:
		pass
	spotlight = MayaSpotlight()
	pos = maya_window().mapToGlobal(QtCore.QPoint(0,0))
	spotlight.setGeometry(pos.x(), pos.y(), maya_window().geometry().width(),200)
	spotlight.show()
	QtGui.QApplication.instance().setActiveWindow(spotlight)
	spotlight.le.setFocus()

def main():
	shortcut = install_shortcut()
	shortcut.start.connect(do_spotlight)